;+
; NAME:
;       COVPM__DEFINE
; PURPOSE:
;       Object class for sequential covariance matrix
;       calculation using the method of provisional means.
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       covpm = Obj_New("COVPM",p)
; ARGUMENTS:
;       p: dimension of the covariance matrix
; KEYWORDS
;       None
; METHODS:
;       UPDATE: update the covariance matrix
;       with an array (spectral tile) of observations
;       covpm -> Update, Xs, weights = Ws
;         Xs is an array of observation row vectors
;         Ws is an optional array of weights for
;         the observations
;       COVARIANCE: return the covariance matrix
;       cov = covpm -> Covariance()
;       MEANS: return the observation means
;       mns = covpm -> Means()
; DEPENDENCIES:
;       PROV_MEANS.DLL
;-------------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

Function COVPM::Init, p
   self.mn = ptr_new(dblarr(p))
   self.cov= ptr_new(dblarr(p,p))
   self.sw = 0.0000001
   return, 1
End

Pro COVPM::Cleanup
   Ptr_Free, self.mn
   Ptr_Free, self.cov
End

Pro COVPM::Update, Xs, weights = Ws
   sz = size(Xs)
   NN = sz[1]
   n = sz[2]
   if n_elements(Ws) eq 0 then Ws = fltarr(sz[2])+1.0
   sw  =  self.sw
   mn  = *self.mn
   cov = *self.cov
   result = call_external('prov_means', 'prov_means', $
       float(Xs), Ws, NN, n, sw, mn, cov, $
       value = [0,0,1,1,0,0,0], /i_value,  /cdecl)
   self.sw   = sw
   *self.mn  = mn
   *self.cov = cov
End

Function COVPM::Covariance
   return, *self.cov/self.sw
End

Function COVPM::Means
   return, *self.mn
End

Pro COVPM__Define
class = {COVPM, $
         sw: 0.0, $      ;current sum of weights
         mn: ptr_new(), $;current mean
         cov: ptr_new()} ;current sum of cross products
End