;+
; NAME:
;       EM
; PURPOSE:
;       FMLE (EM) clustering for Gaussian mixtures.
;       Takes data array Xs (column vectors) and initial
;       class membership probability matrix U as input.
;       Returns U, the class centers Ms, Priors Ps and
;       final class covariances Fs.
;       Allows for simulated annealing
;       Ref: Gath and Geva, IEEE Trans. Pattern Anal. and
;            Mach. Intell. 3(3):773-781, 1989
;            Hilger, Exploratory Analysis of Multivariate
;            Data, PhD Thesis, Tech. Univ. Denmark, 2001
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       EM, Xs, U, Ms, Ps, Ss, $
;        unfrozen=unfrozen,wnd=wnd, $
;        mask=mask, $
;        maxiter=maxiter,miniter=miniter,verbose=verbose,$
;        pdens=pdens,fhv=fhv,T0=T0,$
;        pb_msg=pb_msg,beta=beta, $
;        num_cols=num_cols,num_rows=num_rows
; ARGUMENTS:
;        Xs: input observations array (column vectors)
;        U:  initial class probability membership matrix
;            (column vectors)
;        Ms: cluster means (output)
;        Ps: cluster priors (output)
;        Ss: cluster covariance matrices (output)
; KEYWORDS:
;        unfrozen: Indices of the observations which
;                take part in the iteration (default all)
;        wnd: window for displaying the log likelihood
;             (optional)
;        mask: defines the region to be included (optional)
;        maxinter: maximum iterations (optional)
;        minimter: minimum iterations (optional)
;        pdens: partition density (output, optional)
;        fhv: fuzzy hypervolume (output, optional)
;        T0: initial annealing temperature (default 1.0)
;        verbose: set to print output info to IDL log
;        pb_msg: string for progressbar
;        beta: spatial field parameter
;        num_cols, num_rows: image dimensions (if beta>0)
; DEPENDENCIES:
;       PROGRESSBAR__DEFINE (FSC_COLOR)
;       INTERSECT
;------------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

Pro EM, Xs, U, Ms, Ps, Ss, unfrozen=unfrozen, wnd=wnd, maxiter=maxiter, miniter=miniter, $
        pdens=pdens, fhv=fhv, T0=T0, verbose=verbose, pb_msg=pb_msg, $
        num_cols=num_cols, num_rows=num_rows, beta=beta, mask=mask
V = U*0.0
sz = size(Xs)
n = (size(U))[1]
qf = fltarr(n)
K = (size(U))[2]
if sz[0] eq 1 then NN = 1 else NN = sz[2]
Nb = [[0.0,0.25,0.0],[0.25,0.0,0.25],[0.0,0.25,0.0]]
; check keywords
if n_elements(unfrozen) eq 0 then unfrozen = lindgen(n)
if n_elements(mask) eq 0 then begin
   ismask = 0
   mask = lindgen(n)
end else begin
   unfrozen=intersect(mask,unfrozen)
   ismask = 1
endelse
if n_elements(maxiter) eq 0 then maxiter = 500
if n_elements(miniter) eq 0 then miniter = 10
if n_elements(T0) eq 0 then T0 = 1.0
if n_elements(beta) eq 0 then beta = 0.0
if n_elements(wnd) eq 1 then begin
   wset,wnd
   wxsize=!D.x_size
   wysize=!D.y_size
   if wxsize gt 400 and wysize gt 400 then $
      position=[0.2,0.9-400.0/wysize,0.2+400.0/wxsize,0.9] $
   else position = [0.2,0.1,0.9,0.9]
endif
if n_elements(verbose) eq 0 then verbose = 0
if n_elements(pb_msg) eq 0 then pb_msg='EM: delta_U ...'

if verbose then begin
   print,'EM clustering '+strtrim(n_elements(unfrozen),2)+' pixels'
   print, 'Initial annealing temperature '+strtrim(T0)
endif

; covariance matrix
if NN eq 1 then Ss = dblarr(K) else Ss = dblarr(NN,NN,K)
; vector distances to cluster centers
if NN eq 1 then Ds = dblarr(n) else Ds = dblarr(n,NN)
; work array
if NN eq 1 then W = dblarr(n) else W = dblarr(n,NN)
; log likelihood array
LL = fltarr(maxiter)
; partition density and fuzzy hypervolume arrays
pdens = (fhv = fltarr(K))

; iteration
progressbar = Obj_New('progressbar', Color='blue', Text='0',$
              title= pb_msg,xsize=250,ysize=20)
progressbar->start
dU = 1.0
iter=0L
T = T0
while ((dU gt 0.001) or (iter lt miniter)) and (iter lt maxiter) do begin
  if progressbar->CheckCancel() then begin
     print,'clustering aborted'
     goto, done
  endif
  progressbar->Update,(iter*100)/maxiter,text=strtrim(dU,2)
  Uold = U
; number of (masked) pixels in each cluster
  ns = total(U[mask,*],1)
; prior probabilities
  Ps = ns/n_elements(mask)
; calculate cluster means
; note: A ## TRANSPOSE(B) is equivalent to MATRIX_MULTIPLY(B, A, /ATRANSPOSE)
  Ms = matrix_multiply(U[mask,*], Xs[mask,*], /atranspose)
; loop over the cluster index for the case NN>1
  if NN gt 1 then for j=0,K-1 do begin
    Ms[j,*] = temporary(Ms[j,*])/ns[j]
    for i=0,NN-1 do replicate_inplace,W,Ms[j,i],1,[0,i]
    Ds[mask,*] = Xs[mask,*] - W[mask,*]
;  covariance matrix
    for i=0,NN-1 do W[mask,i] = sqrt(U[mask,j])*Ds[mask,i]
    S = matrix_multiply(W[mask,*],W[mask,*],/atranspose)/ns[j]
    Ss[*,*,j] = S
    sqrtdetS = sqrt(determ(S,/double))
    Sinv = invert(S,/double)
    qf[mask] = total(Ds[mask,*]*(Sinv##Ds[mask,*]),2)
;  class hypervolume and partition density
    fhv[j] = sqrtdetS
    indices = where(qf lt 1.0, count)
    if (count gt 0) then begin
;      if ismask then indices = intersect(mask,indices)
       pdens[j] = total(U[indices,j])/fhv[j]
    endif
;  new memberships
    U[unfrozen,j]=exp(-qf[unfrozen]/2.0)*(Ps[j]/sqrtdetS)
;  random membership for annealing
    if T gt 0.0 then begin
      Ur=1.0-randomu(seed,n_elements(unfrozen))^(1.0/T)
      U[unfrozen,j] = temporary(U[unfrozen,j])*Ur
    endif
  endfor $
; loop over the cluster index for the case NN=1
  else for j=0,K-1 do begin
     Ms[j] = Ms[j]/ns[j]
     replicate_inplace, W, Ms[j]
     Ds[mask] = Xs[mask]-W[mask]
;   variance
     Ss[j] = total(U[mask,j]*Ds[mask]^2)/ns[j]
     sqrtdetS = sqrt(Ss[j])
;   class hypervolume and partition density
     fhv[j] = sqrtdetS
     indices = where(Ds^2/Ss[j] lt 1.0, count)
     if count gt 0 then  begin
        if ismask then indices = intersect(mask,indices)
        pdens[j] = total(U[indices,j])/fhv[j]
     endif
;   new memberships
     U[unfrozen,j] = Ps[j]*exp(-0.5*Ds[unfrozen]^2/Ss[j])/sqrtdetS
;   random membership for annealing
     if T gt 0.0 then begin
       Ur = 1.0 - randomu(seed,n_elements(unfrozen))^(1.0/T)
       U[unfrozen,j] = temporary(U[unfrozen,j])*Ur
     endif
  endfor
; normalize spectral memberships
  a = 1/total(U,2)
  void=where(finite(a),complement=complement,ncomplement=ncomplement)
  if ncomplement gt 0 then a[complement]=0.0
  for j=0,K-1 do U[*,j]=temporary(U[*,j])*a
; determine spatial membership matrix
  if beta gt 0 then begin
    for j=0,K-1 do begin
      U_N=1.0-convol(reform(U[*,j],num_cols, $
                                       num_rows),Nb)
      V[*,j] = exp(-beta*U_N)
    endfor
;  combine spectral/spatial for unfrozen
    U[unfrozen,*]=temporary(U[unfrozen,*])* $
                                      V[unfrozen,*]
;  normalize all
    a = 1/total(U,2)
    void=where(finite(a),complement=complement, $
      ncomplement=ncomplement)
    if ncomplement gt 0 then a[complement]=0.0
    for j=0,K-1 do U[*,j]=temporary(U[*,j])*a
  endif
; log likelihood
  indices=where(U,count)
  if count gt 0 then LL[iter] =  total(Uold[indices]*alog(U[indices]))
  dU = max(abs(U-Uold))
  iter=iter+1
  T = 0.8*T
  if (n_elements(wnd) eq 1) and (iter mod 5 eq 0) then $
    plot, LL[0:iter-1], color=0, background= 'FFFFFF'XL, xtitle='Iteration',position=position

endwhile
done:
Ms = transpose(Ms)
if verbose then begin
   if iter lt maxiter then print, 'converged after '+strtrim(iter,2)+' iterations' $
   else print, 'no convergence after '+strtrim(maxiter,2)+' iterations'
   print,'Partition density for '+strtrim(K,2)+' clusters: ', total(pdens)/total(fhv)
endif
progressbar->destroy


End
