;+
; NAME:
;       EM_RUN
; PURPOSE:
;       ENVI extension for FMLE(EM) clustering at
;       different scales using DWT compression
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       EM_run
; ARGUMENTS:
;       Event (if used as a plug-in menu item)
; KEYWORDS:
;       None
; DEPENDENCIES:
;       ENVI
;       EM (PROGRESSBAR__DEFINE (FSC_COLOR))
;       DWT__DEFINE
;       CLASS_LOOKUP_TABLE
;-------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

pro em_run, event

print, '---------------------------------'
print, 'Scaled EM Clustering'
print, systime(0)
print, '---------------------------------'

;seed = 12345

catch, theError
if theError ne 0 then begin
   void = Dialog_Message(!Error_State.Msg, /error)
   return
endif

envi_select, title='Choose multispectral image', $
             fid=fid, dims=dims,pos=pos,/mask,m_fid=m_fid,m_pos=m_pos
if (fid eq -1) then begin
   print, 'cancelled'
   return
endif

envi_file_query, fid, fname=fname, xstart=xstart, ystart=ystart
print, 'Selected image: ',fname
print, 'Selected bands: ',pos + 1

; number of clusters
base = widget_auto_base(title='Number of Classes')
wg = widget_sslider(base, title='Classes', min=2, max=15, $
  value=6, dt=1, uvalue='slide', /auto)
result = auto_wid_mng(base)
if (result.accept eq 0) then begin
   print, 'cancelled'
   return
endif
K = fix(result.slide)
print, 'Number of Classes', K

; number of compressions
base = widget_auto_base(title='Maximum scaling factor')
wg = widget_sslider(base, title='Num. compressions', min=0, max=3, $
  value=2, dt=1, uvalue='slide', /auto)
result = auto_wid_mng(base)
if (result.accept eq 0) then return
NC = fix(result.slide)
print, 'Number of Compressions', NC

base = widget_auto_base(title='Final scaling factor')
wg = widget_sslider(base, title='Num. compressions', min=0, max=NC, $
  value=0, dt=1, uvalue='slide', /auto)
result = auto_wid_mng(base)
if (result.accept eq 0) then return
NCf = fix(result.slide)
print, 'Final Number of Compressions', NCf

; tie point
map_info = envi_get_map_info(fid=fid)
envi_convert_file_coordinates, fid, dims[1], dims[3], e, n, /to_map
map_info.mc[2:3]= [e,n]

base = widget_auto_base(title='Clustering parameters')
list = ['T0', 'Beta']
vals = [0.5,0.5]
we = widget_edit(base,  list=list, uvalue='edit', $
      vals=vals, field= 2, dt=4, /auto)
result = auto_wid_mng(base)
if (result.accept eq 0) then begin
   print,'cancelled'
   return
endif
T0 = result.edit[0]
beta = result.edit[1]

query=dialog_message('Export clusters to ROIs',/question)
if query eq 'Yes' then roi_flag=1 else roi_flag=0

; output destination
base = widget_auto_base(title='EM Output')
sb = widget_base(base, /row, /frame)
wp = widget_outfm(sb, uvalue='outf', /auto)
result1 = auto_wid_mng(base)
if (result1.accept eq 0) then begin
  print, 'Output cancelled'
  return
endif

base = widget_auto_base(title='Output class membership probs')
sb = widget_base(base, /row, /frame)
wp = widget_outf(sb, uvalue='outf', /auto)
result2 = auto_wid_mng(base)
if (result2.accept eq 0) then begin
  print, 'Output of class membership probs cancelled'
  prob_flag=1
end else prob_flag=0

c_names=strarr(K+1)
fhvs = fltarr(K)
c_names[0]='unclassified'
for i=1,K do c_names[i]='cluster: '+strtrim(i,2)

num_cols = dims[2]-dims[1]+1L
num_rows = dims[4]-dims[3]+1L
num_pixels = (num_cols*num_rows)
num_bands = n_elements(pos)

widget_control, /hourglass

; compress
DWTbands = objarr(num_bands)
for i=0,num_bands-1 do begin
   print,'DWT of band '+strtrim(i+1,2)+'...'
   DWTbands[i] = Obj_New('DWT',envi_get_data(fid=fid,dims=dims,pos=pos[i]))
   for c=1,NC do DWTbands[i]->compress
endfor

; check for mask and scale if present
if (m_fid ne -1) then begin
   DWTmask=Obj_New('DWT',envi_get_data(fid=m_fid,dims=dims,pos=m_pos))
   for c=1,NC do DWTmask->compress
endif

s_num_cols = DWTBands[0]->get_num_cols()
s_num_rows = DWTBands[0]->get_num_rows()
s_num_pixels = s_num_cols*s_num_rows
s_image = fltarr(s_num_pixels,num_bands)
for i=0,num_bands-1 do s_image[*,i] = DWTbands[i]->get_quadrant(0)

; initiailize membership matrix and mask if present
U = randomu(seed,s_num_pixels,K)
for j=0,K-1 do U[*,j]=U[*,j]/total(U,2)
if (m_fid ne -1) then begin
   mask = where(floor(DWTmask->get_quadrant(0)>0) gt 0,count)
   if count eq 0 then mask=lindgen(s_num_pixels)
endif

; initial clustering of maximally compressed image
window,11,xsize=600,ysize=400,title='Log Likelihood'
EM,  s_image, U, Ms, Ps, Ss, T0=T0, wnd=11, $
    pb_msg='EM: Pass 1 of '+strtrim(NC-NCf+1,2), $
    num_cols=s_num_cols, num_rows=s_num_rows, $
    beta=beta, mask=mask

; clustering at increasing scales
for c=1,NC-NCf do begin
; upsample class memberships (bilinear interpolation)
   U = reform(U,s_num_cols,s_num_rows,K,/overwrite)
   U = rebin(U,2*s_num_cols,2*s_num_rows,K)
   s_num_cols = s_num_cols*2
   s_num_rows = s_num_rows*2
   s_num_pixels = s_num_cols*s_num_rows
   U = reform(U,s_num_pixels,K,/overwrite)
; expand the image and mask
   s_image = fltarr(s_num_pixels,num_bands)
   for i=0,num_bands-1 do begin
      DWTBands[i]->expand
      s_image[*,i] = DWTbands[i]->get_quadrant(0)
   endfor
   if (m_fid ne -1) then begin
      DWTmask->expand
      mask = where(floor(DWTmask->get_quadrant(0)>0) gt 0,count)
      if count eq 0 then mask=lindgen(s_num_pixels)
   endif
; cluster
   unfrozen = where(max(U,dimension=2) le 0.9, count)
   if count gt 0 then $
   EM, s_image, U, Ms, Ps, Ss, miniter=5, wnd=11, $
    unfrozen=unfrozen, T0=0.0, pb_msg= $
    'EM: Pass '+strtrim(c+1,2)+' of '+strtrim(NC-NCf+1,2),$
    num_cols=s_num_cols, num_rows=s_num_rows, $
    beta=beta, mask=mask
endfor

print, 'Cluster mean vectors'
print, Ms[*,0:K-1]
print, 'Cluster covariance matrices'
for i=0, K-1 do begin
   print, strtrim(i+1,2)
   if (size(s_image))[0] gt 1 then print, Ss[*,*,i] else print, Ss[i]
endfor

; convert to byte to save memory
U = bytscl(U,min=0.0,max=1.0)

; if image is still downscaled, upsample the class memberships
if NCf gt 0 then  begin
   U = reform(U,s_num_cols,s_num_rows,K,/overwrite)
   s_num_cols=(2^NCf)*s_num_cols
   s_num_rows=(2^NCf)*s_num_rows
   s_num_pixels=s_num_cols*s_num_rows
   U = rebin(U,s_num_cols,s_num_rows,K)
   U = shift(U,2^NCf,2^NCf,0)
   U = reform(U,s_num_pixels,K,/overwrite)
endif

; classify the final expanded image
void = max(U,labels,dimension=2)
if n_elements(labels) eq 1 then goto, done
labels = byte((labels/s_num_pixels)+1)
out_array = bytarr(num_cols,num_rows)
out_array[0:s_num_cols-1,0:s_num_rows-1] = reform(labels,s_num_cols,s_num_rows)
; final mask
if (m_fid ne -1) then begin
   mask_image = envi_get_data(fid=m_fid,dims=dims,pos=m_pos)
   indices=where(mask_image eq 0, count)
   if count gt 0 then out_array[indices]=0
endif

; export clusters to ROIs
if roi_flag then begin
   indices = lindgen(s_num_pixels)
   envi_delete_rois,envi_get_roi_ids()
   for i=1,K do begin
      roi_id = envi_create_roi(color=i,name=c_names[i],ns=num_cols,nl=num_rows)
      ind = where(labels eq i, count)
      if count gt 0 then begin
         xpts = indices[ind] mod s_num_cols
         ypts = indices[ind]/s_num_cols
      endif
      envi_define_roi, roi_id, /point, xpts=xpts, ypts=ypts
   endfor
endif

; output classification image to memory or file
if (result1.outf.in_memory eq 1) then begin
   envi_enter_data, out_array+0, $
                    file_type=3, $
                    map_info=map_info, $
                    xstart=xstart+dims[1], $
                    ystart=ystart+dims[3], $
                    bnames=['FMLE('+fname+')'],$
                    num_classes=K+1, $
                    class_names=c_names, $
                    lookup=class_lookup_table(indgen(K+1))
   print, 'Clusters written to memory'
end else begin
   openw, unit, result1.outf.name, /get_lun
   writeu, unit, out_array
   envi_setup_head ,fname=result1.outf.name, ns=num_cols, nl=num_rows, nb=1, $
                    data_type=1, $
                    interleave=2, $
                    file_type=3, $
                    map_info=map_info, $
                    xstart=xstart+dims[1], $
                    ystart=ystart+dims[3], $
                    bnames=['FMLE('+fname+')'],$
                    num_classes=K+1, $
                    class_names=c_names, $
                    lookup=class_lookup_table(indgen(K+1)), $
                    descrip='scaled EM clustering of ' + result1.outf.name, $
                    /write
   print, 'File created ', result1.outf.name
   free_lun, unit
endelse

if prob_flag then goto, done

; Output membership probabilities to file
openw, unit, result2.outf, /get_lun
for i=0,K-1 do begin
   out_array[0:s_num_cols-1,0:s_num_rows-1,*]=reform(U[*,i],s_num_cols,s_num_rows)
   if (m_fid ne -1) then out_array[indices]=0
   writeu, unit, out_array
endfor
free_lun, unit
envi_setup_head, fname=result2.outf, ns=num_cols, nl=num_rows, nb=K, $
    data_type=1, $
    interleave=0, $
    file_type=0, $
    map_info=map_info, $
    xstart=xstart+dims[1], $
    ystart=ystart+dims[3], $
    bnames='probs: '+c_names, $
    descrip='FMLE probabilities for ' + result2.outf, $
    /write
print, 'File created ', result2.outf

done:
if (m_fid ne -1) then Obj_Destroy,DWTmask
for i=0,num_bands-1 do Obj_Destroy, DWTbands[i]

end