;+
; NAME:
;       FFN__DEFINE
; PURPOSE:
;     Object class for implementation of a
;     two-layer, feed-forward neural network classifier.
;     This is a generic class with no training methods.
; AUTHOR
;     Mort Canty (2006)
;     Juelich Research Center
;     m.canty@fz-juelich.de
; CALLING SEQUENCE:
;     ffn = Obj_New("FFN",Xs,Ls,L)
; ARGUMENTS:
;     Xs:  array of observation column vectors
;     Ls:  array of class label column vectors
;          of form (0,0,1,0,0,...0)^T
;     L:   number of hidden neurons
; KEYWORDS
;     None
; METHODS (external):
;     FORWARDPASS: propagates a biased input
;                  observation vector through
;                  the network. Returns the softmax
;                  probabilities vector, sets the
;                  hidden layer outputs as side effect
;       m = ffn -> ForwardPass()
;     VFORWARDPASS: propagates an array of biased input
;                   vectors through the network,
;                   returns the corresponding softmax
;                   probability vectors and hidden
;                   layer outputs in variable Ns
;       Ms = ffn -> vForwardPass(Ns)
;     CLASSIFY: returns the classes for an array
;               of observation vectors X
;               returns the class probabilities in
;               array variable PROBS
;       c = ffn -> Classify(X,Probs)
;     COST: returns the current cross entropy
;       c = ffn -> Cost(key)
;       key = 0 for all training data,
;             1 for odd-numbered,
;             2 for even numbered
;     GET_WEIGHTS: return synaptic weights as a vector
;       w = ffn -> Get_Weights()
;     put_WEIGHTS: insert synaptic weights into network
;       ffn -> Put_Weights, w
; DEPENDENCIES:
;     None
;------------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

Function FFN::Init, Xs, Ls, L
   catch, theError
   if theError ne 0 then begin
      catch, /cancel
      ok = dialog_message(!Error_State.Msg + ' Returning...', /error)
      return, 0
   endif
; network architecture
   self.LL = L
   self.np = n_elements(Xs[*,0])
   self.NN = n_elements(Xs[0,*])
   self.KK = n_elements(Ls[0,*])
; biased output vector from hidden layer (column vector)
   self.N= ptr_new(fltarr(L+1))
; biased exemplars (column vectors)
   self.Xs = ptr_new([[fltarr(self.np)+1],[Xs]])
   self.Ls = ptr_new(Ls)
; weight matrices (each column is a neuron weight vector)
   self.Wh = ptr_new(randomu(seed,L,self.NN+1)-0.5)
   self.Wo = ptr_new(randomu(seed,self.KK,L+1)-0.5)
   return,1
End

Pro FFN::Cleanup
   ptr_free, self.Xs
   ptr_free, self.Ls
   ptr_free, self.Wh
   ptr_free, self.Wo
   ptr_free, self.N
End

Function FFN::forwardPass, x
; logistic activation for hidden neurons, N set as side effect
   *self.N = [[1],[1/(1+exp(-transpose(*self.Wh)##x))]]
; softmax activation for output neurons
   I = transpose(*self.Wo)##*self.N
   A = exp(I-max(I))
   return, A/total(A)
End

Function FFN::vForwardPass, Xs, N
; vectorized forward pass, returns output vector
; N is set as a side effect
   np = n_elements(Xs[*,0])
   Ones = fltarr(np) + 1.0
   N = [[Ones],[1/(1+exp(-transpose(*self.Wh)##Xs))]]
   Io = transpose(*self.Wo)##N
   maxIo = max(Io,dimension=2)
   for k=0,self.KK-1 do Io[*,k]=Io[*,k]-maxIo
   A = exp(Io)
   sum = total(A,2)
   M = fltarr(np,self.KK)
   for k=0,self.KK-1 do M[*,k] = A[*,k]/sum
   return, M
End

Function FFN:: classify, X, Probs
; vectorized class membership probabilities
   nx = n_elements(X[*,0])
   Ones = fltarr(nx) + 1.0
   N = [[Ones],[1/(1+exp(-transpose(*self.Wh)##[[Ones],[X]]))]]
   Io = transpose(*self.Wo)##N
   maxIo = max(Io,dimension=2)
   for k=0,self.KK-1 do Io[*,k]=Io[*,k]-maxIo
   A = exp(Io)
   sum = total(A,2)
   Probs = fltarr(nx,self.KK)
   for k=0,self.KK-1 do Probs[*,k] = A[*,k]/sum
; vectorized class memberships
   maxM = max(Probs,dimension=2)
   M=fltarr(self.KK,nx)
   for i=0,self.KK-1 do M[i,*]=Probs[*,i]-maxM
   return, byte((where(M eq 0.0) mod self.KK)+1)
End

Function FFN:: cost, key
   case key of
      0: indices = lindgen(self.np)                                           ; all training data
      1: void = where( (lindgen(self.np) mod 2) eq 0, complement=indices )    ; odd-numbered training data
      2: indices = where( (lindgen(self.np) mod 2) eq 0 )                     ; even-numbered training data
   endcase
   np = n_elements(indices)
   Xs=(*self.Xs)[indices,*]
   Ls=(*self.Ls)[indices,*]
   Ones = fltarr(np) + 1.0
   N = [[Ones],[1/(1+exp(-transpose(*self.Wh)##[Xs]))]]
   Io = transpose(*self.Wo)##N
   maxIo = max(Io,dimension=2)
   for k=0,self.KK-1 do Io[*,k]=Io[*,k]-maxIo
   A = exp(Io)
   sum = total(A,2)
   Ms = fltarr(np,self.KK)
   for k=0,self.KK-1 do Ms[*,k] = A[*,k]/sum
   return, -total(Ls*alog(Ms+1e-40))
End


Function FFN::GetWeights
   return, [(*self.Wh)[*],(*self.Wo)[*]]
End

Pro FFN::PutWeights, w
   *self.Wh = reform(w[0:self.LL*(self.NN+1)-1],self.LL,self.NN+1)
   *self.Wo = reform(w[self.LL*(self.NN+1):*],self.KK,self.LL+1)
End

Pro FFN__Define
class =  { FFN, $
           NN: 0L,               $ ;input dimension
           LL: 0L,               $ ;number of hidden units
           KK: 0L,               $ ;output dimension
           Wh:ptr_new(),         $ ;hidden weights
           Wo:ptr_new(),         $ ;output weights
           Xs:ptr_new(),         $ ;training pairs
           Ls:ptr_new(),         $
           N:ptr_new(),          $ ;output vector from hidden layer
           np: 0L                $ ;number of training pairs
          }
End