;+
; NAME:
;       FKM
; PURPOSE:
;       Fuzzy Kmeans clustering algorithm.
;       Takes data array Xs and initial fuzzy
;       membership matrix (as column vectors).
;       Returns fuzzy membership matrix U and
;       the class centers Ms.
;       Ref: J. C. Dunn, Journal of Cybernetics,
;       (1973) PAM1:32-57
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       FKM, Xs, K, U, Ms, niter=niter, seed=seed
; ARGUMENTS:
;        Xs: input observations array (column vectors)
;        U:  class probability membership matrix
;            (input and output)
;        Ms: cluster means (output)
; KEYWORDS:
;        niter: number of iterations (optional)
;        unfrozen: observations taking part in
;        iteration (default all)
; DEPENDENCIES:
;       PROGRESSBAR__DEFINE (FSC_COLOR)
;-----------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

pro FKM, Xs, U, Ms, niter=niter, unfrozen=unfrozen

n = n_elements(Xs[*,0])
NN = n_elements(Xs[0,*])
K = (size(U))[2]

if n_elements(niter) eq 0 then niter = 500
if n_elements(unfrozen) eq 0 then unfrozen = lindgen(n)

; vector distances to cluster centers
Ds = fltarr(n,NN)
; work array
W = fltarr(n,NN)

; iteration
progressbar = Obj_New('progressbar', Color='blue', Text='0',$
              title='FKM clustering: delta_U...',xsize=250,ysize=20)
progressbar->start
dU = 1.0 & iter=0L
while ((dU gt 0.001) or (iter lt 20)) and (iter lt niter) do begin
   if progressbar->CheckCancel() then begin
      print,'clustering aborted'
      progressbar->Destroy
      return
   endif
   progressbar->Update,(iter*100)/niter,text=strtrim(dU,2)
   Uold = U
   UU = U*U
   ; update means and distances
   Ms = Xs ## transpose(UU)
   for j=0,K-1 do begin
      Ms[j,*]=Ms[j,*]/total(UU[*,j])
      for i=0,NN-1 do W[*,i]=replicate(Ms[j,i],n)
      Ds = Xs-W
      dd = 1/total(Ds*Ds,2)
      U[unfrozen,j] = dd[unfrozen]
   endfor
; normalize
   for j=0,K-1 do U[*,j]=U[*,j]/total(U,2)
   dU = max(abs(U-Uold))
   iter=iter+1
endwhile

Ms = transpose(Ms)
progressbar->destroy

end
