;+
; NAME:
;       HCL
; PURPOSE:
;       Agglomerative hierarchic clustering with
;       sum of squares cost function.
;       Takes data array Xs (column vectors) and number
;       of clusters K as input.
;       Returns cluster memberships Cs.
;       Ref. Fraley Technical Report 311, Dept. of Stat.,
;       University of Washington, Seattle (1996).
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       HCL, Xs, K, Cs
; ARGUMENTS:
;        Xs: input observations array (column vectors)
;        K:  number of clusters
;        Cs: Cluster labels of observations
; KEYWORDS:
;        None
; DEPENDENCIES:
;       PROGRESSBAR__DEFINE (FSC_COLOR)
;----------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

function best_merge, D
; return the best merge, larger index first
   n = n_elements(D[*,0])
   pos= (where(D eq min(D)))[0]
   return, [pos mod n, pos/n]
end

pro hcl, Xs, KK, Cs

n = n_elements(Xs[0,*]) ; dimensionality of data
m = n_elements(Xs[*,0]) ; number of sample vectors
mults = lonarr(m)+1     ; initial multiplicities
Cs = findgen(m)         ; initial cluster labels

; initialize merge-cost array
Delta=fltarr(m,m)+10e10
for j=1,m-1 do begin
    for i=0,j-1 do begin
       Delta[j,i] = total((Xs[i,*]-Xs[j,*])^2)/2
    endfor
endfor

; begin iteration
cost = 0.0
c = m
progressbar = Obj_New('progressbar', Color='blue', Text='0',$
              title='HCL: classes remaining...',xsize=250,ysize=20)
progressbar->start
while c gt KK do begin
   if progressbar->CheckCancel() then begin
      print,'clustering aborted'
      Cs=-1
      progressbar->Destroy
      return
   endif
   progressbar->Update,(m-c)*100/m,text=strtrim(c,2)
   bm = best_merge(Delta)
   j = bm[0]
   i = bm[1]
   cost = cost + Delta[j,i]
   for k=0,m-1 do if Cs[k] eq j then Cs[k] = i
   for k=0,m-1 do if Cs[k] gt j then Cs[k] = Cs[k]-1
; pre-merge multiplicities
   ni = mults[i]
   nj = mults[j]
; update merge-cost array
   for k=i+1,c-1 do begin
      nk = mults[k]
      Djk = 0
      if j gt k then Djk = Delta[j,k] else if j lt k then Djk = Delta[k,j]
      Delta[k,i] = ( (ni+nk)*Delta[k,i]+(nj+nk)*Djk-nk*Delta[j,i] )/(ni+nj+nk)
   endfor
   for k=0,i-1 do begin
      nk = mults[k]
      Djk = 0
      if j gt k then Djk = Delta[j,k] else if j lt k then Djk = Delta[k,j]
      Delta[i,k] = ( (ni+nk)*Delta[i,k]+(nj+nk)*Djk-nk*Delta[j,i] )/(ni+nj+nk)
   endfor
; update multiplicities
   mults[i] = mults[i]+mults[j]
; delete the upper cluster
   if j eq c-1 then begin
      mults = mults[0:j-1]
      Delta = Delta[0:j-1,*]
      Delta = Delta[*,0:j-1]
   end else begin
      mults = [mults[0:j-1],mults[j+1:*]]
      Delta = [Delta[0:j-1,*],Delta[j+1:*,*]]
      Delta = transpose( [ transpose(Delta[*,0:j-1]),transpose(Delta[*,j+1:*]) ] )
   end
   c = c-1
endwhile
progressbar->destroy
end
