;+
; NAME:
;       KERNEL_RUN
; PURPOSE:
;       ENVI extension for classification of
;       a multispectral image with a kernel-based
;       classifier
;       Ref: T. Masters, Advanced Algorithms for
;       Neural Networks (1995) Wiley
; AUTHOR;
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       Kernel_Run
; ARGUMENTS
;       Event (if used as a plug-in menu item)
; KEYWORDS:
;       None
; DEPENDENCIES:
;       ENVI
;       MINF_BRACKET
;       MINF_PARABOLIC (PROGRESSBAR_DEFINE FSC_COLOR)
;----------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

Function Output, X, sigma
; returns the class membership probability array
   common exemplars, Xs, Ls, maxLs, W, K, NN, n
   result = fltarr(K)
   for i=0,NN-1 do W[*,i]=X[i]
   d = Xs - W
   d2 = total(d*d,2)
; remove training exemplar X, if present
   d2 = d2[where(d2 gt 0)]/(sigma*sigma)
   exps = exp(-d2)
   for j=0,K-1 do begin
      indices = where(maxLs/n eq j,nj)
      result[j] = total(exps[indices])/nj
   endfor
   return, result/total(result)
End

Function Theta, sigma
; returns the current misclassification rate
   common exemplars, Xs, Ls, maxLs, W, K, NN, n
   cnt = 0
   for i=0L,n-1 do begin
; get training exemplar
      X = Xs[i,*]
      L = Ls[i,*]
; get its class
      void = max(output(X,sigma),k1)
; and its label
      void = max(L,k2)
; misclassification?
      if k1 ne k2 then cnt++
   endfor
   result = float(cnt)/n
   oplot,[sigma],[result],psym=5,color=0
   return, result
End

Pro kernel_run, event
   common exemplars, Xs, Ls, maxLs, W, K, NN, n
   print, '---------------------------------'
   print, 'Kernel-based Classification'
   print, systime(0)
   print, '---------------------------------'
; select the image to be classified
   envi_select, title='Enter File for classification', fid=fid, pos=pos, dims=dims
   if (fid eq -1) then begin
      print,'cancelled'
      return
   endif
   envi_file_query, fid, fname=fname,xstart=xstart, ystart=ystart, interleave=interleave
   if interleave eq 0 then begin
      answer = dialog_message(file_basename(fname)+' will be coverted to BIP. Continue?',/question)
      if answer eq 'No' then begin
         print,'cancelled'
         return
      endif
      dims =  [-1L,0,ns-1,0,nl-1]
      pos = lindgen(nb)
      ENVI_DOIT, 'CONVERT_INPLACE_DOIT', fid=fid, o_interleave=2, dims=dims, pos=pos, r_fid=r_fid
      fid = r_fid
      interleave = 2
   endif
   print, 'file: '+fname
   print, 'bands: ',pos
   num_cols=dims[2]-dims[1]+1
   num_rows=dims[4]-dims[3]+1
   n_bands=n_elements(pos)
; tie point
   map_info = envi_get_map_info(fid=fid)
   envi_convert_file_coordinates, fid, dims[1], dims[3], e, n, /to_map
   map_info.mc[2:3]= [e,n]

; get associated ROIs
   roi_ids = envi_get_roi_ids(fid=fid, roi_names=roi_names, roi_colors=roi_colors)
   if (roi_ids[0] eq -1) then begin
      error=dialog_message('No ROIs associated with the selected file',/error)
      print, 'No ROIs associated with the selected file'
      print, 'done'
      return
   endif
; compound widget for ROI selection
   base = widget_auto_base(title='ROI Selection')
   wm   = widget_multi(base, list=roi_names, uvalue='list', /auto)
   result = auto_wid_mng(base)
   if (result.accept eq 0) then begin
      error=dialog_message('No ROIs selected',/error)
      print, 'No ROIs selected'
      print, 'done'
      return
   endif
   ptr = where(result.list eq 1, K)

; output destination
   base = widget_auto_base(title='Output kernel-based classification to file')
   sb = widget_base(base, /row, /frame)
   wp = widget_outf(sb, uvalue='outf', /auto)
   result = auto_wid_mng(base)
   if (result.accept eq 0) then begin
     print, 'output of classification aborted'
     return
   endif
   class_filename=result.outf
   openw, class_unit, class_filename, /get_lun
   base = widget_auto_base(title='Output probability image to file')
   sb = widget_base(base, /row, /frame)
   wp = widget_outf(sb, uvalue='outf', /auto)
   result = auto_wid_mng(base)
   if (result.accept eq 0) then begin
      print, 'output of probability image cancelled'
      prob_filename='none'
      pr_flag=0
   end else begin
      prob_filename=result.outf
      pr_flag=1
      openw, prob_unit, prob_filename, /get_lun
   endelse

; output file for test results
  test_flag = 1
  outfile=dialog_pickfile(filter='*.tst',/write,/overwrite_prompt,title='Save test results to disk')
  if (outfile eq '') then begin
     print,'output of test results cancelled'
     test_flag = 0
  endif

; construct the exemplar matrices
   Xs = fltarr(n_elements(pos))
   Ls = fltarr(K)
   for i=0,K-1 do begin
      Xs1 = envi_get_roi_data(roi_ids[ptr[i]], fid=fid, pos=pos)
      Xs  = [[Xs], [Xs1]]
      Ls1 = fltarr(K,n_elements(Xs1[0,*]))
      Ls1[i,*] = 1.0
      Ls = [[Ls], [Ls1]]
   endfor
   Xs = transpose(Xs)
   Ls = transpose(Ls)
   Xs = Xs[1:n_elements(Xs[*,0])-1,*]
   Ls = Ls[1:n_elements(Ls[*,0])-1,*]
   NN = n_elements(Xs[0,*])
   n = n_elements(Xs[*,0])

   if test_flag then begin
; split into training and test data
      test_indices = where( (lindgen(n) mod 3) eq 0, $
                     complement = train_indices )
      Xs_test = Xs[test_indices,*]
      Ls_test = Ls[test_indices,*]
      n_test = n_elements(Xs_test[*,0])
      n = n - n_test
      Xs = Xs[train_indices,*]
      Ls = Ls[train_indices,*]
   endif

   print, 'Number of training pixels is '+strtrim(n,2)

; train the classifier

; set variables W and maxLs in common block
   W = fltarr(n,NN)
   void = max(Ls,maxLs,dimension=2)

   window,12,xsize=600,ysize=400, $
     title='Kernel-based classification: bracketing ...'
   wset, 12
   widget_control, /hourglass

; bracket the minimum
   s1=0.1  & s2=10.0
   minF_bracket, s1,s2,s3, FUNC_NAME="Theta"

   window,12,xsize=600,ysize=400, $
     title='Kernel-based classification: iterating ...'
   plot,[s1,s3],[theta(s1),theta(s3)], $
      psym=5,color=0,background='FFFFFF'XL, $
      xtitle = textoidl('\sigma'), $
      ytitle = textoidl('\theta')

; hunt it down
   minF_parabolic, s1,s2,s3, sigma, theta_min, $
                FUNC_NAME="Theta"

   if test_flag then begin
; test the classifier
      test_classes = bytarr(n_test)
      for i=0,n_test-1 do begin
         void = max(output(Xs_test[i,*],sigma),index)
         test_classes[i] = index+1
      endfor
      openw,lun,outfile,/get_lun
      printf,lun,'; FFN test results for '+ fname
         printf,lun,'; '+systime(0)
         printf,lun,'; Classification image: '+ class_filename
         printf,lun,'; Class probabilities image: '+ prob_filename
         printf,lun, n_test
         void = max(Ls_test,labels,dimension=2)
         labels = byte((labels/n_test)+1)
         printf,lun, transpose([[test_classes], [labels]])
         free_lun,lun
         print, 'test results written to '+outfile
   endif

; classify the image
   progressbar = Obj_New('progressbar', Color='blue', Text='0',$
                 title='Classifying tile ...',xsize=250,ysize=20)
   progressbar->start
   print,'classifying...'
; get (spectral) tile id
   tile_id = envi_init_tile(fid,pos,num_tiles=num_tiles, $
       interleave=interleave,xs=dims[1],xe=dims[2],ys=dims[3],ye=dims[4])
   x = fltarr(1,n_bands)
   class_row = bytarr(num_cols)
   prob_row = bytarr(K,num_cols)
; start (spectral) tiling
   for tile_index=0L,num_tiles-1 do begin
      if interleave eq 2 then tile = transpose(envi_get_tile(tile_id,tile_index)) $
                         else tile = envi_get_tile(tile_id,tile_index)
      if progressbar->CheckCancel() then begin
         print,'Classification aborted'
         free_lun, class_unit
         if pr_flag eq 1 then free_lun, prob_unit
         progressbar->Destroy
         return
      endif
      progressbar->Update,tile_index*100/num_tiles,text=strtrim(tile_index,2)
      for i=0,num_cols-1 do begin
         pv = output(tile[i,*],sigma)
         void = max(pv,class)
         class_row[i] = class+1
         prob_row[*,i] = bytscl(pv,min=0.0,max=1.0)
      endfor
      writeu,class_unit,class_row
      if pr_flag eq 1 then $
 ; write probabilities in BIP, byte format
         writeu,prob_unit,prob_row
   endfor ; end  of tiling
   progressbar->destroy
   envi_tile_done, tile_id
   lookup = [[0,0,0],[roi_colors[*,ptr]]]
   envi_setup_head,fname=class_filename, ns=num_cols, $
                   nl=num_rows, nb=1, $
                   data_type=1, $
                   file_type=3, $
                   map_info=map_info,$
                   num_classes=K+1, $
                   interleave=0, /write, /open, $
                   class_names=['unclassified',roi_names[ptr]], $
                   xstart=xstart, $
                   ystart=ystart, $
                   lookup = lookup, $
                   bnames=['KERNEL('+fname+')']
   print, 'classification file created ', class_filename
   free_lun, class_unit
   if pr_flag eq 1 then begin
      envi_setup_head, fname=prob_filename, ns=num_cols, $
                   nl=num_rows, nb=K, $
                   map_info=map_info,$
                   data_type=1, $
                   interleave=2, $ ; BIP
                   /write, /open, $
                   xstart=xstart, $
                   ystart=ystart, $
                   bnames=roi_names[ptr]
      print, 'probabilities file created ', prob_filename
      free_lun, prob_unit
   endif
   print, 'done'

End