;+
; NAME:
;       MAD_ITER
; PURPOSE:
;       Function for Iteratively Re-weighted Multivariate
;       Alteration Detection (IR-MAD).
;       Ref: A. A. Nielsen et al. Remote Sensing
;            of Environment 64 (1998), 1-19
;            A. A. Nielsen submitted for publication
;       Input files must be BIL or BIP format and
;       have equal spatial/spectral dimensions.
;       On error or if interrupted during the first
;       iteration, returns = -1 else 0
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       result = Mad_Iter(fid1,fid2,dims1,dims2,pos1,pos2,$
;         A=A,B=B,NCP=NCP,chi_sqr=chi_sqr,niter=niter,
;         means1=means1, means2=means2, rho=rho)
; ARGUMENTS:
;       fid1, fid2      input file specifications
;       dims1, dims2
;       pos1, pos2
; KEYWORDS:
;       A, B           transformation eigenvectors,
;                      smallest correlation first
;       means1,means2  weighted mean values, row-replicated
;       rho            canonical correlations in decreasing
;                      order
;       ncp            no change probabilities image
;       chi_sqr        chi-square image
;       niter          max number of iterations,
;                      default is 50
;       verbose        print intermediate results to log
;                      default 0 (false)
; DEPENDENCIES:
;       ENVI
;       COVPM_DEFINE
;       GEN_EIGENPROBLEM
;       DWT__DEFINE
;       PROGRESSBAR_DEFINE (FSC_COLOR)
;---------------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

Function mad_iter, fid1, fid2, dims1, dims2, pos1, pos2, niter=niter, rho=rho, $
               NCP=NCP,CHI_SQR=CHI_SQR, A=A, B=B, means1=means1, means2=means2, $
               verbose=verbose

   if n_elements(niter) eq 0 then niter = 50
   if n_elements(verbose) eq 0 then verbose = 0
   if N_elements(mad_std) eq 0 then mad_std = 0

   envi_file_query, fid1, interleave=interleave1
   envi_file_query, fid2, interleave=interleave2

   if (interleave1 eq 0) or (interleave2 eq 0) then begin
      print, 'MAD_ITER: nput image is BSQ, aborting'
      return, -1
   endif

   num_cols = dims1[2]-dims1[1]+1
   num_rows = dims1[4]-dims1[3]+1
   num_pixels = (num_cols*num_rows)
   num_bands = n_elements(pos1)

   if niter gt 0 then rhos=fltarr(num_bands,niter)-1

   if niter gt 1 then begin
      window, 12, xsize=600, ysize=400, title='IR-MAD'
      wset,12
      plot, indgen(niter), rhos[0,*], psym=4, xrange = [0,niter+1], yrange = [0,1.1], $
         xtitle='iteration', ytitle='Correlations',color=0,background='FFFFFF'XL
   endif

; get tile ids
   tile_id1 = envi_init_tile(fid1,pos1,num_tiles=num_tiles, $
       interleave=interleave1,xs=dims1[1],xe=dims1[2],ys=dims1[3],ye=dims1[4])
   tile_id2 = envi_init_tile(fid2,pos2, $
       interleave=interleave2,xs=dims2[1],xe=dims2[2],ys=dims2[3],ye=dims2[4])

; arrays
   NCP  = fltarr(num_cols,num_tiles)+1.0
   means1 = fltarr(num_bands,num_cols)
   means2 = fltarr(num_bands,num_cols)
   sdevMADs = fltarr(num_bands)

; begin iteration
   iter=0L
   delta = 1.0
   old_rho = fltarr(num_bands)
   repeat begin
     MADs = fltarr(num_bands,num_cols,num_rows)
     progressbar = Obj_New('progressbar', Color='blue', Text=' ',$
                    title='Iter '+strtrim(iter,2),xsize=300,ysize=20)
     progressbar->start
     covpm = obj_new('COVPM',2*num_bands)
     txt = 'Pass 1 (statistics)...'
; spectral tiling for statistics
     for tile_index=0L,num_tiles-1 do begin
        if progressbar->CheckCancel() then begin
           if iter gt 0 then begin
              iter = niter
              txt = 'Interrupting...'
           end else begin
              print,'Calculation aborted'
              obj_destroy, covpm
              progressbar->destroy
              wdelete,12
              return, -1
           endelse
        endif
        if tile_index mod 10 eq 0 then begin
           pct = (tile_index)*100/(num_tiles)
           progressbar->Update,fix(pct) ,text=txt
        endif
        tile1 = envi_get_tile(tile_id1,tile_index)
        tile2 = envi_get_tile(tile_id2,tile_index)
        if interleave1 eq 1 then tile1 = transpose(tile1)
        if interleave2 eq 1 then tile2 = transpose(tile2)
        covpm->update,[tile1,tile2],weights=NCP[*,tile_index]

     endfor
     progressbar->destroy
; canonical correlation
     Sigma = covpm->Covariance()
     means = covpm->Means()

     Obj_Destroy, covpm

     for i=0,num_cols-1 do begin
        means1[*,i]=means[0:num_bands-1]
        means2[*,i]=means[num_bands:2*num_bands-1]
     endfor
     S_xx = Sigma[0:num_bands-1,0:num_bands-1]
     S_yy = Sigma[num_bands:2*num_bands-1,num_bands:2*num_bands-1]
     S_xy = Sigma[num_bands:2*num_bands-1,0:num_bands-1]
     S_yx = Sigma[0:num_bands-1,num_bands:2*num_bands-1]
     C1 = S_xy ## invert(S_yy,/double) ## S_yx
     B1 = S_xx
     C2 = S_yx ## invert(S_xx,/double) ## S_xy
     B2 = S_yy
     gen_eigenproblem, C1, B1, A, lambda ; C1 a = lambda B1 a
     gen_eigenproblem, C2, B2, B, lambda ; C2 b = lambda B2 b

     rho = sqrt(lambda)
     sigMads = sqrt(2*(1-rho))

; stopping criterion

     delta = max(abs(rho-old_rho))
     old_rho = rho
     if verbose then print, 'delta = '+strtrim(delta,2)

; the variances of the canonical variates U = A'X and V = B'Y  are unity
; ensure sum of positive correlations between X and U is positive
; their covariance matrix is S_xx##A
     invsig_x = diag_matrix(1/sqrt(diag_matrix(S_xx)))
     sum = total(invsig_x##S_xx##A,2)
     A = A##diag_matrix(sum/abs(sum))

; ensure positive correlation between each pair of canonical variates
     cov = diag_matrix(transpose(A)##S_xy##B)
     B = B##diag_matrix(cov/abs(cov))

     if (iter ne 0) and (iter eq niter) then goto, done

; spectral tiling to determine MADs for chi-square
     progressbar = Obj_New('progressbar', Color='blue', Text=' ',$
                    title='Iter '+strtrim(iter,2),xsize=300,ysize=20)
     progressbar->start
     txt = 'Pass 2 (collecting)... '
     for tile_index=0L,num_tiles-1 do begin
         if progressbar->CheckCancel() then begin
            if iter gt 0 then begin
               txt = 'Interrupting...'
               iter = niter
            end else begin
               print,'Calculation aborted'
               progressbar->destroy
               wdelete,12
               envi_tile_done, tile_id1
               envi_tile_done, tile_id2
               return, -1
            endelse
         endif
         if tile_index mod 10 eq 0 then begin
            pct = (tile_index)*100/(num_tiles)
            progressbar->Update,fix(pct) ,text=txt
         endif
         tile1 = envi_get_tile(tile_id1,tile_index)
         tile2 = envi_get_tile(tile_id2,tile_index)
         if interleave1 eq 1 then tile1 = transpose(tile1)
         if interleave2 eq 1 then tile2 = transpose(tile2)
         MADs[*,*,tile_index]  = (tile1-means1)##A-(tile2-means2)##B
     endfor
     progressbar->destroy

     MADs = reform(MADS,num_bands,num_pixels,/overwrite)
     for i=0,num_bands-1 do MADs[i,*] = MADs[i,*]/sigMADs[i]

     chi_sqr = reform( total((MADs)^2,1), num_cols, num_tiles )
     NCP = 1.0 - chisqr_pdf(chi_sqr,num_bands)

     if (iter ne 0) and (iter eq niter) then goto, done

;    print sigmas and plot canonical correlations
     if verbose then begin
        print, 'iteration '+strtrim(iter,2)
        print, reverse(sigMADs), format='("Sigma MADs",'+strtrim(num_bands,2)+'f10.5)'
     endif
     if niter gt 1 then begin
        rhos[*,iter] = rho
        wset,12
        plot, indgen(iter+1)+1, rhos[0,0:iter], psym=-4, xrange=[0,niter+1],  yrange = [0,1.1], $
           xtitle='iteration', ytitle='Correlations',color=0,background='FFFFFF'XL
        for i=1,num_bands-1 do $
          oplot, indgen(iter+1)+1, rhos[i,0:iter], psym=-4, color=0
     endif
  done:
      iter=iter+1
  endrep until (iter gt niter) or (delta lt 0.001)

; successfully finished, so
  envi_tile_done, tile_id1
  envi_tile_done, tile_id2
return, 0 & End