;+
; NAME:
;       MAXLIKE_RUN
; PURPOSE:
;       ENVI extension for classification of
;       a multispectral image with maximum likelihood
; AUTHOR;
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       Maxlike_Run
; ARGUMENTS
;       Event (if used as a plug-in menu item)
; KEYWORDS:
;       None
; DEPENDENCIES:
;       ENVI
;----------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

Pro maxlike_run, event
   print, '---------------------------------'
   print, 'Maximum Likelihood Supervised Classification'
   print, systime(0)
   print, '---------------------------------'
; select the image to be classified
   envi_select, title='Enter File for classification', fid=fid, pos=pos, dims=dims
   if (fid eq -1) then begin
      print,'cancelled'
      return
   endif
   envi_file_query, fid, fname=fname, xstart=xstart, ystart=ystart
   print, 'file: '+fname
   num_cols=dims[2]-dims[1]+1
   num_rows=dims[4]-dims[3]+1
   n_bands=n_elements(pos)
; tie point
   map_info = envi_get_map_info(fid=fid)
   envi_convert_file_coordinates, fid, dims[1], dims[3], e, n, /to_map
   map_info.mc[2:3]= [e,n]

; get associated ROIs
   roi_ids = envi_get_roi_ids(fid=fid, roi_names=roi_names, roi_colors=roi_colors)
   if (roi_ids[0] eq -1) then begin
      error=dialog_message('No ROIs associated with the selected file',/error)
      print, 'No ROIs associated with the selected file'
      print, 'done'
      return
   endif
; compound widget for ROI selection
   base = widget_auto_base(title='ROI Selection')
   wm   = widget_multi(base, list=roi_names, uvalue='list', /auto)
   result = auto_wid_mng(base)
   if (result.accept eq 0) then begin
      error=dialog_message('No ROIs selected',/error)
      print, 'No ROIs selected'
      print, 'done'
      return
   endif
   ptr = where(result.list eq 1, K)

; output destination
   base = widget_auto_base(title='Output MaxLike classification to file')
   sb = widget_base(base, /row, /frame)
   wp = widget_outf(sb, uvalue='outf', /auto)
   result = auto_wid_mng(base)
   if (result.accept eq 0) then begin
     print, 'output of classification aborted'
     return
   endif
   class_filename=result.outf
   openw, class_unit, class_filename, /get_lun
   base = widget_auto_base(title='Output MaxLike rule image to file')
   sb = widget_base(base, /row, /frame)
   wp = widget_outf(sb, uvalue='outf', /auto)
   result = auto_wid_mng(base)
   if (result.accept eq 0) then begin
      print, 'output of rule image cancelled'
      rule_filename='none'
      rule_flag=0
   end else begin
      rule_filename=result.outf
      rule_flag=1
      openw, rule_unit, rule_filename, /get_lun
   endelse

; output file for test results
  test_flag = 1
  outfile=dialog_pickfile(filter='*.tst',/write,/overwrite_prompt,title='Save test results to disk')
  if (outfile eq '') then begin
     print,'output of test results cancelled'
     test_flag = 0
  endif

; construct the exemplar matrices for MaxLike
   Xs = fltarr(n_elements(pos))
   Ls = fltarr(K)
   for i=0,K-1 do begin
      Xs1 = envi_get_roi_data(roi_ids[ptr[i]], fid=fid, pos=pos)
      Xs  = [[Xs], [Xs1]]
      Ls1 = fltarr(K,n_elements(Xs1[0,*]))
      for j=0,n_elements(Xs1[0,*])-1 do Ls1[i,j]=1.0
      Ls = [[Ls], [Ls1]]
   endfor
   Xs = transpose(Xs)
   Ls = transpose(Ls)
   n = n_elements(Xs[*,0])
   Xs = Xs[1:n_elements(Xs[*,0])-1,*]
   Ls = Ls[1:n_elements(Ls[*,0])-1,*]

   if test_flag then begin
; split into training and test data
      test_indices = where( (lindgen(n) mod 3) eq 0, $
                     complement = train_indices )
      Xs_test = Xs[test_indices,*]
      Ls_test = Ls[test_indices,*]
      n_test = n_elements(Xs_test[*,0])
      n = n - n_test
      Xs = Xs[train_indices,*]
      Ls = Ls[train_indices,*]
   endif

; train the classifier
   mn =  fltarr(n_bands,K)
   cov = fltarr(n_bands,n_bands,K)
   class_names = strarr(K+1)
   class_names[0]='unclassified'
   void = max(Ls,labels,dimension=2)
   labels = byte((labels/n))
   for i=0,K-1 do begin
      class_names[i+1]='class'+string(i+1)
      indices = where(labels eq i,count)
      if count gt 1 then begin
         XXs = transpose(Xs[indices,*])
         for j=0,n_bands-1 do mn[j,i] = mean(XXs[j,*])
         cov[*,*,i] = correlate(XXs,/covariance)
      endif
   endfor

envi_check_save, /classification

   if rule_flag then begin
       envi_doit, 'class_doit', fid=fid, dims=dims, $
            pos=pos, out_bname='MaxLike('+fname+')', $
            class_names=class_names, $
            lookup=class_lookup_table(indgen(K+1)+1), $
            method=2, mean=mn, cov=cov, $
            out_name = class_filename, $
            rule_out_name = rule_filename, $
            rule_out_bname=roi_names[ptr]
       print, 'classification file created ', class_filename
       print, 'rule file created ', rule_filename
   end  else  begin
      envi_doit, 'class_doit', fid=fid, dims=dims, $
            pos=pos, out_bname='MaxLike('+fname+')', $
            class_names=class_names, $
            lookup=class_lookup_table(indgen(K+1)+1), $
            method=2, mean=mn, cov=cov, $
            out_name = class_filename
      print, 'classification file created ', class_filename
   endelse

   if test_flag then begin
; test the classifier
      dummy = fltarr(1,n_test,n_bands)
      dummy[0,*,*] = Xs_test
      dummy_dims = [-1,0,0,0,n_test-1]
      envi_enter_data, dummy, r_fid=r_fid
      envi_doit, 'class_doit', fid=r_fid, r_fid=r_fid1, dims=dummy_dims, pos=pos, out_bname='Test', $
            class_names=class_names, lookup=class_lookup_table(indgen(K+1)+1), $
            method=2, mean=mn, cov=cov, /in_memory
      test_classes = (envi_get_data(fid=r_fid1,pos=0,dims=dummy_dims))[*]
      envi_file_mng, id=r_fid,/remove
      envi_file_mng, id=r_fid1,/remove
      openw,lun,outfile,/get_lun
      printf,lun,'; MaxLike test results for '+ fname
      printf,lun,'; '+systime(0)
      printf,lun,'; Classification image: '+ class_filename
      printf,lun,'; Class rule image: '+ rule_filename
      printf,lun, n_test, K
      void = max(Ls_test,labels,dimension=2)
      labels = byte((labels/n_test)+1)
      printf,lun, transpose([[test_classes], [labels]])
      free_lun,lun
      print, 'test results written to ' + outfile
   endif

End