;+
; NAME:
;       ORTHO_REGRESS
; PURPOSE:
;       Orthogonal regression between two vectors
;       Ref: M. Canty et al. Remote Sensing of Environment
;           91(3,4) (2004) 441-451
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       Ortho_Regress, X, Y, b, Xm, Ym, sigma_a, sigma_b
;       regression line is Y = Ym + b(X-Xm)
;                             = (Ym-bXm) + bX = a + bX
; ARGUMENTS:
;       input column vectors X and Y
;       returns b, Xm, Ym, sigma_a, sigma_b, sigma(RMSE)
; KEYWORDS:
;       None
; DEPENDENCIES:
;       None
;-------------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

pro ortho_regress, X, Y, b, Xm, Ym, $
                   sigma_a, sigma_b, sigma
   m = n_elements(X)
   Xm = mean(X)
   Ym = mean(Y)
   S = correlate([X,Y],/covariance,/double)
   Sxx = S[0,0]
   Syy = S[1,1]
   Sxy = S[1,0]
   void = eigenql(S,eigenvectors=eigenvectors,/double)
; slope
   b = eigenvectors[1,0]/eigenvectors[0,0]
; standard errors
   sigma2 = m*(Syy-2*b*Sxy+b*b*Sxx)/((m-2)*(1+b^2))
   tau = sigma2*b/((1+b^2)*Sxy)
   sigma_b=sqrt( sigma2*b*(1+b^2)*(1+tau)/(m*Sxy))
   sigma_a=sqrt( sigma2*b*(1+b^2)*(Xm^2*(1+tau)+Sxy/b)$
                                          /(m*Sxy) )
   sigma = sqrt(sigma2)
end