;+
; NAME:
;       WISHART
; PURPOSE:
;       Test two multivariate distributions for
;       equal means and covariance matrices
;       assuming both parameter sets are estimated
;       with the same sample size
;       Ref. Anderson (2003) pp. 416-426
; AUTHOR:
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       pVal=Wishart(M1,M2,S1,S2,n,statistic=statistic)
;       returns the probability of observing a
;       larger value of the realization of the
;       test statistic Q = -2*rho*log(lambda)
;       under the null hypothesis
; ARGUMENTS:
;       M1, M2 are the means
;       S1, S2 are the covariance matrices
;       n is the (common) number of samples
;       used to estimate them
; KEYWORDS:
;       statistic
;       the test statistic -2*rho*log(lambda)
; DEPENDENCIES
;       None
;-----------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

FUNCTION Wishart, M1, M2, S1, S2, n, statistic=statistic
   NN = n_elements(M1)
   B = n*(S1 + S2 + 0.5*transpose(M1-M2)##(M1-M2))
   logLambda = n*(NN*alog(2*n)+0.5*alog(determ(S1))+0.5*alog(determ(S2))-alog(determ(B)))
   rho = 1.0-1.5*(2.0*NN^2+9.0*NN+11.0)/(n*(6.0*(NN+3.0)))
   w2  = NN*(NN+3.0)*1.5*((NN+1)*(NN+2)-6.0*(1.0-rho^2))/(n^2*48.0*rho^2)
   df  = NN*(NN+1)/2
   statistic = -2.0*rho*logLambda
   Pr1 = chisqr_pdf(statistic,df)
   Pr2 = chisqr_pdf(statistic,df+4)
   return, 1-(Pr1+w2*(Pr2-Pr1))
END