;+
; NAME:
;       ATWT_ADD_RUN
; PURPOSE:
;       ENVI extension for panchromatic sharpening under
;       ARSIS model with "A trous" wavelet transform.
;       Ref: Aiazzi et al, IEEE Transactions on Geoscience
;       and Remote Sensing, 40(10) 2300-2312, 2002
;       Uses additive model
;       Ref Nunez et al. IEEE Transactions on Geoscience
;            and Remote Sensing 37(3) (1999) 1204-1211
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       ATWT_ADD_RUN
; ARGUMENTS:
;       Event (if used as a plug-in menu item)
; KEYWORDS:
;       None
; DEPENDENCIES:
;       ENVI
;       ATWT__DEFINE
;       ORTHO_REGRESS
;------------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

Pro atwt_add_run, event

   print, '-----------------------------------'
   print, 'ARSIS ATWT Model for Wavelet Fusion'
   print, systime(0)
   print, '-----------------------------------'

catch, theError
if theError ne 0 then begin
   catch,/cancel
   ok = dialog_message(!Error_State.msg,/error)
   return
endif

; get MS image
   envi_select, title='Select  low resolution multi-band input file', $
                       fid=fid, dims=dims, pos=pos
   if (fid eq -1) then begin
     print,'cancelled'
     return
   endif
   print,'Dimensions of low resolution spatial subset'
   print, dims[1:4]
   num_cols = dims[2]-dims[1]+1L
   num_rows = dims[4]-dims[3]+1L
   num_bands = n_elements(pos)
   envi_file_query, fid, fname=fname,bnames=bnames,data_type=data_type,xstart=xstart,ystart=ystart
   map_info = envi_get_map_info(fid=fid)
   ps = map_info.ps[0]
   print,'Low resolution (multispectral) image chosen: ',fname
   print,'Bands ',pos+1

; get pan image
   envi_select, title='Select high resolution (pan) input band',fid=fid1,pos=pos1,dims=dims1,/band
   if (fid1 eq -1) then begin
     print,'cancelled'
     return
   endif
   envi_file_query, fid1, fname=fname1,data_type=data_type1,xstart=xstart1,ystart=ystart1
   map_info1 = envi_get_map_info(fid=fid1)
; get number of applications necessary
   ps1 = map_info1.ps[0]
   num_trans = round(ps/(2*ps1))
   ratio = round(ps/ps1)
   print,'High resolution image chosen: ',fname1
   print,'Band ',pos1+1

; get output destination
   base = widget_auto_base(title='ARSIS Fusion Output')
   sb = widget_base(base, /row, /frame)
   wp = widget_outfm(sb, uvalue='outf', /auto)
   result = auto_wid_mng(base)
   if (result.accept eq 0) then begin
     print, 'cancelled'
     return
   endif

   widget_control,/hourglass

; find upper left position of low-res image within pan image
   envi_convert_file_coordinates,fid,dims[1],dims[3],e,n,/to_map
   envi_convert_file_coordinates,fid1,X_ul,Y_ul,e,n
   X_ul = round(X_ul)
   Y_ul = round(Y_ul)
; shift to pixel center
   if num_trans eq 2 then begin
      X_ul = X_ul-num_trans
      Y_ul = Y_ul-num_trans
   endif

; cutout the corresponding spatial subset of the pan image
   num_cols1 = (2^num_trans)*num_cols
   num_rows1 = (2^num_trans)*num_rows

   dimsp = [-1L,X_ul,X_ul+num_cols1-1,Y_ul,Y_ul+num_rows1-1]
   if (dimsp[1] lt dims1[1]) or (dimsp[3] lt dims1[3]) or $
      (dimsp[2] gt dims1[2]) or (dimsp[4] gt dims1[4]) then message,'PAN dims out of bounds'

   pan = envi_get_data(fid=fid1,dims=dimsp,pos=pos1)
   print,'Dimensions of pan subset'
   print,dimsp[1:4]+1
   str=[file_basename(fname1),strtrim(dimsp[1]+1)+strtrim(dimsp[2]+1)+strtrim(dimsp[3]+1)+strtrim(dimsp[4]+1)]
   envi_info_wid, str, title='Pan spatial subset'

; tie point for fused image
   envi_convert_file_coordinates, fid1, dimsp[1], dimsp[3], e, n, /to_map
   map_info1.mc[2:3]= [e,n]

; output arrays for fusion
   case data_type of
    1: image_fused = bytarr(num_cols1,num_rows1,num_bands)
    2: image_fused = intarr(num_cols1,num_rows1,num_bands)
    4: image_fused = fltarr(num_cols1,num_rows1,num_bands)
   12: image_fused = uintarr(num_cols1,num_rows1,num_bands)
   else: begin
         void = dialog_message('format must be byte, float or (un)signed integer',/error)
         return
       end
   endcase
   smpl = randomu(seed,100000L,/long) mod (num_cols1*num_rows1)

; loop over MS bands
   for k=0,num_bands-1 do begin
; transform the pan image
      msATWT = Obj_New('ATWT',pan)
      for i=1,num_trans do msATWT->compress
; resize the ms band to scale of the pan image using NN resampling
      ms_band=envi_get_data(fid=fid,dims=dims,pos=pos[k])
      envi_enter_data, ms_band,r_fid=r_fid
      dims2 = [-1L,0,num_cols-1,0,num_rows-1]
      envi_doit, 'resize_doit',/in_memory,fid=r_fid,pos=0,dims=dims2,r_fid=r_fid1,interp=0,rfact=[1.0,1.0]/2^num_trans
      dims3 = [-1L,0,num_cols1-1,0,num_rows1-1]
      ms_band=envi_get_data(fid=r_fid1,dims=dims3,pos=0)
      envi_file_mng, id=r_fid,/remove
      envi_file_mng, id=r_fid1,/remove
; add pan details
      msATWT->inject,ms_band
; expand to full resolution
      for i=1,num_trans do msATWT->expand
; prepare for output
      temp = msATWT->get_image(0)
      temp = temp > 0.0
      case data_type of
        1: image_fused[*,*,k] = byte(temp < 255.0)
        2: image_fused[*,*,k] = fix(temp < 2047.0)
        4: image_fused[*,*,k] = float(temp)
       12: image_fused[*,*,k] = uint(temp < 2047.0)
      endcase
      Obj_destroy, msATWT
   endfor ; end loop over bands

; output to memory or file in BSQ
   if (result.outf.in_memory eq 1) then begin
      envi_enter_data, image_fused,map_info=map_info1,xstart=xstart+dims[1], ystart=ystart+dims[3], $
                       bnames='ARSIS_ATWT:'+bnames[pos],$
                       descrip='ARSIS_ATWT: high-res='+file_basename(fname)+' low-res(ms)='+file_basename(fname1)
      print, 'Result written to memory'
   endif else begin
      openw, unit, result.outf.name, /get_lun
      for i=0,num_bands-1 do writeu, unit, image_fused[*,*,i]
      envi_setup_head ,fname=result.outf.name, ns=num_cols1, $
           nl=num_rows1, nb=num_bands, $
           map_info=map_info1, $
           xstart=xstart+dims[1], ystart=ystart+dims[1], $
           data_type=data_type, interleave=0, /write, $
           bnames='ARSIS_ATWT:'+bnames[pos],$
           /open,$
           descrip='ARSIS_ATWT: high-res='+file_basename(fname)+' low-res(ms)='+file_basename(fname1)
      print, 'File created ', result.outf.name
      free_lun, unit
   endelse

End

