;+
; NAME:
;       EKM_RUN
; PURPOSE:
;       ENVI extension for extended K-means clustering
; AUTHOR
;       Mort Canty (2006)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       EKM_Run
; ARGUMENTS:
;       Event (if used as a plug-in menu item)
; KEYWORDS:
;       None
; DEPENDENCIES:
;       ENVI
;       HCL (PROGRESSBAR__DEFINE (FSC_COLOR))
;       CLASS_LOOKUP_TABLE
;----------------------------------------------------
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
;----------------------------------------------------------

pro EKM_run, event

print, '---------------------------------'
print, 'Extended K-means Clustering'
print, systime(0)
print, '---------------------------------'

n_samples = 500

;catch, theError
;if theError ne 0 then begin
;   void = Dialog_Message(!Error_State.Msg, /error)
;   return
;endif

envi_select, title='Choose multispectral image for clustering', $
             fid=fid, dims=dims,pos=pos
if (fid eq -1) then begin
   print, 'cancelled'
   return
endif

envi_file_query, fid, fname=fname, xstart=xstart, ystart=ystart
print, 'Selected image: ',fname
print, 'Selected bands: ',pos + 1

; tie point
map_info = envi_get_map_info(fid=fid)
envi_convert_file_coordinates, fid, dims[1], dims[3], e, n, /to_map
map_info.mc[2:3]= [e,n]

; number of clusters (meta-parameter)
base = widget_auto_base(title='Approximate Number of Classes')
wg = widget_sslider(base, title='Classes', min=2, max=15, $
  value=6, dt=1, uvalue='slide', /auto)
result = auto_wid_mng(base)
if (result.accept eq 0) then begin
   print, 'cancelled'
   return
endif
Ktilde = byte(result.slide)
print, 'Cluster meta-parameter', Ktilde

; output destination
base = widget_auto_base(title='EKM Output')
sb = widget_base(base, /row, /frame)
wp = widget_outfm(sb, uvalue='outf', /auto)
result = auto_wid_mng(base)
if (result.accept eq 0) then begin
  print, 'cancelled'
  return
endif

num_cols = dims[2]-dims[1]+1
num_rows = dims[4]-dims[3]+1
num_pixels = (num_cols*num_rows)
num_bands = n_elements(pos)

; read image data, sample for HCL algorithm and set meta-parameters
indices = round( randomu(seed,n_samples)*num_pixels ) < num_pixels
indices = randomu(seed,n_samples,/long) mod num_pixels
samples = fltarr(n_elements(indices),num_bands)
image = fltarr(num_pixels,num_bands)
sigma2= 0.0
for i=0,num_bands-1 do begin
   temp=float(envi_get_data(fid=fid,dims=dims,pos=pos[i]))
   samples[*,i]=temp[indices]
   image[*,i] = temp
   sigma2=sigma2+variance(temp-shift(temp,[1,0]))
endfor
sigma2 = sigma2/num_bands
alphaE = -1/(2*alog(1.0/Ktilde))

means = fltarr(50,num_bands)
Dds = fltarr(num_pixels) + 10e10
ms  = fltarr(num_pixels)
logps = fltarr(num_pixels)
work = fltarr(num_pixels)
labels = bytarr(num_pixels)
priors = fltarr(50)

; run HCL to get an initial clustering with 50 clusters
print, 'Hierarchic clustering algorithm running on ',n_elements(samples[*,0]),' pixels ...'
HCL, samples, 50, Cs

; generate an initial array of labels (0..49) for each pixel
; by minimum distance to cluster center
progressbar = Obj_New('progressbar', Color='blue', $
  title='HCL clustering: k ...',xsize=250,ysize=20)
progressbar->start
for i = 0,49 do begin
   if progressbar->CheckCancel() then begin
      print,'clustering aborted'
      progressbar->Destroy & return
   endif
   progressbar->Update,(i*100)/49, $
                  text=strtrim(i,2)
   indices = where(Cs eq i)
   means[i,*] = total(samples[indices,*],1)/n_elements(indices)
   work = work*0.0
   for j = 0,num_bands-1 do begin
      replicate_inplace,ms,means[i,j]
      work = work + (image[*,j]-ms)^2
   endfor
   Dds1 = work
   indices = where(Dds1 lt Dds, count)
   if count gt 0 then begin
      labels[indices] = i+1
      Dds[indices] = Dds1[indices]
   endif
endfor
progressbar->Destroy

; EKM clustering
print, 'Extended K-means clustering algorithm running on ',num_pixels,' pixels ...'
for i=0,19 do begin
   void = where(labels eq i, ni)
   priors[i] = float(ni)/num_pixels
endfor
progressbar = Obj_New('progressbar', Color='blue', $
  title='EKM clustering: K (delta) ...',xsize=250,ysize=20)
progressbar->start
delta = 100.0 & iter = 0 & K=50
while (delta gt 0.001) and (iter le 500) do begin
   if progressbar->CheckCancel() then begin
      print,'clustering aborted'
      iter=500
   endif
; drop (almost) empty clusters
   indices = where(priors gt 0.01, K)
   progressbar->Update,(iter*100)/500, $
                  text=strtrim(K,2)+' ('+strtrim(delta,2)+')'
   means = means[indices,*]
   priors = priors[indices]  & print,priors
   dds = fltarr(num_pixels,K)
   means1 = means
   means = means*0.0
; re-cluster
   for k=0,K-1 do begin
      work = work*0.0
      for j = 0,num_bands-1 do begin
         replicate_inplace,ms,means1[k,j]
         work = temporary(work) + (image[*,j]-ms)^2
      endfor
      replicate_inplace,logps,alog(priors[k])
      dds[*,k] = work/(2*sigma2) - alphaE*logps
   endfor
   min_ds = min(dds,dimension=2)
   for k=0,K-1 do begin
      indices = where(dds[*,k] eq min_ds,count)
      if count gt 0 then begin
         nk = n_elements(indices)
         priors[k] = float(nk)/num_pixels
         means[k,*] = total(image[indices,*],1)/nk
         labels[indices] = k+1
      endif
   endfor
   delta = max(abs(means - means1))
iter++
endwhile
progressbar->Destroy

c_names = strarr(K+1)
c_names[0] = 'unclassified'
for i=1,K do c_names[i]='cluster: '+strtrim(i,2)

; write result to memory or disk
for i=1,K do c_names[i]='cluster: '+strtrim(i,2)
if (result.outf.in_memory eq 1) then begin
   envi_enter_data, byte(reform(labels,num_cols,num_rows)), $
                    file_type=3, $
                    map_info=map_info, $
                    bnames=['EKM('+fname+')'],$
                    num_classes=K+1, $
                    class_names=c_names, $
                    xstart=xstart+dims[1], $
                    ystart=ystart+dims[3], $
                    lookup=class_lookup_table(indgen(K+1))
   print, 'Result written to memory'
endif else begin
   openw, unit, result.outf.name, /get_lun
   writeu, unit, byte(reform(labels,num_cols,num_rows))
   envi_setup_head ,fname=result.outf.name, ns=num_cols, nl=num_rows, nb=1, $
       data_type=1, $
       interleave=2, $
       file_type=3, $
       map_info=map_info, $
       bnames=['EKM('+fname+')'],$
       num_classes=K+1, $
       class_names=c_names, $
       xstart=xstart+dims[1], $
       ystart=ystart+dims[3], $
       lookup=class_lookup_table(indgen(K+1)), $
       descrip='EKM clustering of ' + result.outf.name, $
       /write
   print, 'File created ', result.outf.name
   free_lun, unit
endelse

end


