;+
; NAME:
;       MAF_RUN
; PURPOSE:
;       ENVI extension for Min/Max Autocorrelation Factor transformation.
;       Ref: Green et al, IEEE Transaction on Geoscience and Remote Sensing, 26(1):65-74,1988
;       No tiling, so not suitable for large images
; AUTHOR
;       Mort Canty (2004)
;       Juelich Research Center
;       m.canty@fz-juelich.de
; CALLING SEQUENCE:
;       Maf_Run
; ARGUMENTS:
;       Event (if used as a plug-in menu item)
; KEYWORDS:
;       None
; DEPENDENCIES:
;       ENVI
;       GEN_EIGENPROBLEM
;---------------------------------------------------------------------

pro maf_run, event

print, '------------------------------------'
print, 'Maximum Autocorrelation Factor (MAF)'
print, systime(0)
print, '------------------------------------'

catch, theError
if theError ne 0 then begin
   void = Dialog_Message(!Error_State.Msg, /error)
   return
endif

envi_select, title='Choose multispectral image', fid=fid, dims=dims,pos=pos
if (fid eq -1) then begin
   print, 'Canceled'
   return
endif

if n_elements(pos) lt 2 then begin
   print, 'Aborted'
   Message, 'Spectral subset size must be at least 2'
endif

base = widget_auto_base(title='MAF Output')
sb = widget_base(base, /row, /frame)
wp = widget_outfm(sb, uvalue='outf', /auto)
result = auto_wid_mng(base)
if (result.accept eq 0) then begin
  print, 'Cancelled'
  return
endif

; tie point
map_info = envi_get_map_info(fid=fid)
envi_convert_file_coordinates, fid, dims[1], dims[3], e, n, /to_map
map_info.mc[2:3]= [e,n]

envi_file_query, fid, fname=fname, xstart=xstart, ystart=ystart, descrip=descrip

widget_control,/hourglass

num_cols = dims[2]-dims[1]+1
num_rows = dims[4]-dims[3]+1
num_pixels = (num_cols*num_rows)
num_bands = n_elements(pos)

image=fltarr(num_bands,num_pixels)

samples1=fltarr(num_bands,num_pixels)
samples2=fltarr(num_bands,num_pixels)
temp=fltarr(num_pixels)
for i=0,num_bands-1 do begin
   temp=envi_get_data(fid=fid,dims=dims,pos=pos[i])
   m=mean(temp)
   image[i,*]=temp-m
   image1=shift(image[i,*],1,0)
   samples1[i,*]=image[i,*]-image1
   image1=shift(image[i,*],0,1)
   samples2[i,*]=image[i,*]-image1
endfor
sigma =  correlate(image,/covariance,/double)
sigma1 = (correlate(samples1,/covariance,/double)+correlate(samples2,/covariance,/double))/2

gen_eigenproblem, sigma1, sigma, A, lambda
print, 'Eigenvalues:'
print, reverse(lambda)
print, 'Signal to noise'
print, 2/reverse(lambda)
; eigenvectors A are scaled so that MAFs have unit variance.
; ensure sum of positive correlations between X and MAF is positive
; their covariance matrix is cov_xm = Sigma##A
invsig_x = diag_matrix(1/sqrt(diag_matrix(Sigma)))
; correlation matrix
corr_xm = invsig_x##Sigma##A
; adjust sign of eigenvectors
sum = total(corr_xm,2)
A = A##diag_matrix(sum/abs(sum))

MAFs = image ## A

MAF_array = fltarr(num_cols,num_rows,num_bands)
; MAF variances are all equal to 1
sigMAFs = fltarr(num_bands)+1.0
for i = 0,num_bands-1 do MAF_array[*,*,num_bands-i-1] = reform(MAFs[i,*],num_cols,num_rows,/overwrite)

band_names=strarr(num_bands)
for i=0,num_bands-1 do band_names[i]='MAF ' + strtrim(i+1,2)
if (result.outf.in_memory eq 1) then begin
   envi_enter_data, MAF_array, $
   map_info=map_info, $
   xstart=xstart+dims[1],ystart=ystart+dims[3],$
   bnames=band_names, $
   descrip='MAF: '+file_basename(fname)
   print, 'Result written to memory'
endif else begin
   openw, unit, result.outf.name, /get_lun
   writeu, unit, MAF_array
   envi_setup_head ,fname=result.outf.name, ns=num_cols, nl=num_rows, nb=num_bands, $
       data_type=4, interleave=0, /write, /open, $
	   bnames=band_names, $
	   map_info=map_info, $
	   xstart=xstart+dims[1],ystart=ystart+dims[3], $
	   descrip='MAF: '+file_basename(fname)
   print, 'File created ', result.outf.name
   free_lun, unit
endelse

end